#Original C source: Copyright (C) Peter B.L. Meijer
#http://www.seeingwithsound.com/im2sound.htm
#Check site for license details.

import pyglet
#disables wglShareLists context sharing and shadowwindow to prevent fatal crash
pyglet.options['shadow_window'] = False
from pyglet.window import key

from pyglet.media.drivers.openal import lib_openal as al
from pyglet.media.drivers.openal import lib_alc as alc

import numpy as np
import struct
import ctypes
import math


#main class
class example(pyglet.window.Window):
    def __init__(self):
        super(example, self).__init__(640, 480, resizable=False, fullscreen=False, caption="Example")
        self.clear()

        self.key_input = []

    #load listener
        self.listener = Listener()
    #load our custom OpenAL sound buffer
        self.buffer = buffer_sound()
    #load sonifier
        self.sonifier = sonifier()
    #load sound player
        self.player = Player()

    #set listener position
        self.listener.position = (0,240,0)

    #load sample image
        self.image = pyglet.image.load('test.png')

    #sonify toggle
        self.scan_toggle = False
    #loop toggle
        self.loop = False

    #temporary frame buffer image
        self.tmp_image = None

        pyglet.clock.get_fps()
        pyglet.clock.schedule_interval(self.update, .01)



#main update loop
    def update(self,dt):
    #play sound
        if 'SPACE press' in self.key_input:
            if self.scan_toggle == False:
                self.scan_toggle = True
                print 'playing'

    #repeat toggle
        elif 'RETURN press' in self.key_input:
            if self.loop == False:
                self.loop = True
                print 'loop enabled'
            else:
                self.loop = False
                print 'loop disabled'

    #increase speed
        elif 'MINUS press' in self.key_input:
            self.sonifier.rate(False)
            print 'speed up',self.sonifier.speed

    #decrease speed
        elif 'EQUAL press' in self.key_input:
            self.sonifier.rate(True)
            print 'slow down',self.sonifier.speed

    #process image
        if self.tmp_image != None and self.sonifier.output == None:# and self.scan_toggle == True: #> 5
            self.sonifier.generate(self.tmp_image)
            self.tmp_image = None
            print 'processing'

    #purge/load new sound data
        elif self.player.playing() == False and self.sonifier.output != None:# and self.scan_toggle == False:
            if len(self.player.queue) > 0:
                self.player.remove()
            self.buffer.load(self.sonifier.output)
            self.player.add(self.buffer)
            self.player.play()
            self.sonifier.output = None
            if self.loop == True:
                self.scan_toggle = True
            print 'playing'

    #Move player position for stereo panning effect
        if self.player.playing():
            self.player.position = [-100+(self.player.seek*100),240,50]

    #clear key buffer
        if len(self.key_input) > 0:
            self.key_input = []

    #draw screen
        self.draw()


#draw scene
    def draw(self):
        self.clear()
        self.image.blit(0,0)
        if self.scan_toggle == True:
            self.tmp_image = pyglet.image.get_buffer_manager().get_color_buffer().get_region(0,0,64,64).get_texture()
            self.scan_toggle = False

       
#key press input
    def on_key_press(self,symbol,modifiers):
        self.key_input.append(key.symbol_string(symbol) + " press")
    #quit
        if symbol == key.ESCAPE:
            self.player.remove()
            self.player.delete()
            self.buffer.delete()
            self.listener.delete()
            self.close()


#Key Release Input
    def on_key_release(self,symbol,modifiers):
        self.key_input.append(key.symbol_string(symbol) + " release")



#image to sound converter class
class sonifier(object):
#pre-calculate waveform values
    def __init__(self):
        self.speed = 1.0
    #sample rate/length of our audio data
        self.byte = int(21000 * self.speed)
    #size of target image(s)
        self.size = 64

        self.ns = self.byte
        self.scale = 0.5/math.sqrt(self.size)
        self.TwoPi = 6.283185307179586476925287 #double Pi
        self.rnd = 0.211321159122
        self.phi0 = self.TwoPi * self.rnd

    #Generate sine wave to multiply image data against
        self.w = np.asarray(range(self.size))
        self.w = self.TwoPi * 500 * np.power(1.0*5000/500, 1.0*self.w/(self.size-1))
        self.w = np.tile(self.w,(self.ns,1))

        self.core = np.asarray(range(self.ns)) #312.5 * data(4096), repeat
        self.core = np.reshape(self.core,(self.byte,1))
        self.core = np.repeat(self.core,(self.size),axis=1)
    #finished sine wave
        self.w = np.sin(self.w * (self.core * (1.0 / self.byte)) + self.phi0)
    #initialize final output ctypes array
        self.output = None#(ctypes.c_short * self.ns)()

#reset array sizes for altered speed
    def rate(self,speed):
        if speed == False and self.speed > 0.025:
            self.speed -= 0.025
        elif speed == True and self.speed < 1.975:
            self.speed += 0.025
        self.byte = int(21000 * self.speed)
        self.ns = self.byte
    #recalculate sine wave
        self.w = np.asarray(range(self.size))
        self.w = self.TwoPi * 500 * np.power(1.0*5000/500, 1.0*self.w/(self.size-1))
        self.w = np.tile(self.w,(self.ns,1))

        self.core = np.asarray(range(self.ns)) #312.5 * data(4096), repeat
        self.core = np.reshape(self.core,(self.byte,1))
        self.core = np.repeat(self.core,(self.size),axis=1)
    #finished sine wave
        self.w = np.sin(self.w * (self.core * (1.0 / self.byte)) + self.phi0)
    #reinitialize final output ctypes array
        self.output = None#(ctypes.c_short * self.ns)()


#load and convert image, multiply it by waveform and load into ctypes array for ouput
    def generate(self,image):
        self.image = image
        rawimage = self.image.get_image_data()
        pixels = rawimage.get_data('RGBA',rawimage.width*4)
    #convert image data to numpy array int16
        self.data = np.fromstring(pixels, dtype='uint8')

    #reshape array into 4 by (width*height) array, then strip color data except red channel.
        self.data = np.reshape(self.data,(len(self.data)/4,4))
        self.data = self.data[::,:1:]

    #reshape array into X by Y grid to match image X/Y
        self.data = np.reshape(self.data,(self.size,self.size))
    #convert image data to floating point.
        self.data = self.data / 255.0
    #expand image data to match size of wav sample
        self.data = np.repeat(self.data,int(np.ceil(self.byte / float(self.size))),axis=1) 
        self.data = self.data[::,:self.byte:]


        self.data = np.swapaxes(self.data,0,1)

    #multiply array for volume adjustment
        self.data = np.multiply(self.data,self.w) * 1500.0

    #sum array data into 1D array, creating our final waveform
        self.data = np.sum(self.data,axis=1)

    #add click tone to beginning
        self.data[:64] = ((2.0 * self.rnd - 1.0) / self.scale) * 32767.0

    #load data into a ctypes array for dumping into OpenAL buffer
        self.output = (ctypes.c_short * self.ns)()
        for a in xrange(0,len(self.output),1):
            self.output[a] = self.output[a]+int(self.data[a])


#OpenAL Sound Buffer
class buffer_sound(object):
    def __init__(self):
        self.channels = 1
        self.bitrate = 16
        self.samplerate = 21000
        self.wavbuf = None
        self.alformat = al.AL_FORMAT_MONO16
        self.length = None
##        formatmap = {
##            (1, 8) : al.AL_FORMAT_MONO8,
##            (2, 8) : al.AL_FORMAT_STEREO8,
##            (1, 16): al.AL_FORMAT_MONO16,
##            (2, 16) : al.AL_FORMAT_STEREO16,
##        }
##        alformat = formatmap[(channels, bitrate)]


        self.buf = al.ALuint(0)
        al.alGenBuffers(1, self.buf)

    def load(self,data):
        self.wavbuf = data
        self.length = len(data)
    #allocate buffer space to: buffer, format, data, len(data), and samplerate
        al.alBufferData(self.buf, self.alformat, self.wavbuf, len(self.wavbuf)*2, self.samplerate)

    def delete(self):
        al.alDeleteBuffers(1, self.buf)



#load a listener to load and play sounds.
class Listener(object):
    def __init__(self):
    #load device/context/listener
        self.device = alc.alcOpenDevice(None)
        self.context = alc.alcCreateContext(self.device, None)
        alc.alcMakeContextCurrent(self.context)

#set player position
    def _set_position(self,pos):
        self._position = pos
        x,y,z = map(int, pos)
        al.alListener3f(al.AL_POSITION, x, y, z)

    def _get_position(self):
        return self._position

#delete current listener
    def delete(self):
        alc.alcDestroyContext(self.context)
        alc.alcCloseDevice(self.device)

    position = property(_get_position, _set_position,doc="""get/set position""")



#Sound Player class
class Player(object):
#load default settings
    def __init__(self):
    #load source player
        self.source = al.ALuint(0)
        al.alGenSources(1, self.source)
    #disable rolloff factor by default
        al.alSourcef(self.source, al.AL_ROLLOFF_FACTOR, 0)
    #disable source relative by default
        al.alSourcei(self.source, al.AL_SOURCE_RELATIVE,0)
    #capture player state buffer
        self.state = al.ALint(0)
    #set internal variable tracking
        self._volume = 1.0
        self._pitch = 1.0
        self._position = [0,0,0]
        self._rolloff = 1.0
        self._loop = False
        self.queue = []


#set rolloff factor, determines volume based on distance from listener
    def _set_rolloff(self,value):
        self._rolloff = value
        al.alSourcef(self.source, al.AL_ROLLOFF_FACTOR, value)

    def _get_rolloff(self):
        return self._rolloff


#set whether looping or not - true/false 1/0
    def _set_loop(self,lo):
        self._loop = lo
        al.alSourcei(self.source, al.AL_LOOPING, lo)

    def _get_loop(self):
        return self._loop
      

#set player position
    def _set_position(self,pos):
        self._position = pos
        x,y,z = map(int, pos)
        al.alSource3f(self.source, al.AL_POSITION, x, y, z)

    def _get_position(self):
        return self._position
        

#set pitch - 1.5-0.5 float range only
    def _set_pitch(self,pit):
        self._pitch = pit
        al.alSourcef(self.source, al.AL_PITCH, pit)

    def _get_pitch(self):
        return self._pitch

#set volume - 1.0 float range only
    def _set_volume(self,vol):
        self._volume = vol
        al.alSourcef(self.source, al.AL_GAIN, vol)

    def _get_volume(self):
        return self._volume

#queue a sound buffer
    def add(self,sound):
        al.alSourceQueueBuffers(self.source, 1, sound.buf) #self.buf
        self.queue.append(sound)

#remove a sound from the queue (detach & unqueue to properly remove)
    def remove(self):
        if len(self.queue) > 0:
            al.alSourceUnqueueBuffers(self.source, 1, self.queue[0].buf) #self.buf
            al.alSourcei(self.source, al.AL_BUFFER, 0)
            self.queue.pop(0)

#play sound source
    def play(self):
        al.alSourcePlay(self.source)

#get current playing state
    def playing(self):
        al.alGetSourcei(self.source, al.AL_SOURCE_STATE, self.state)
        if self.state.value == al.AL_PLAYING:
            return True
        else:
            return False

#stop playing sound
    def stop(self):
        al.alSourceStop(self.source)

#rewind player
    def rewind(self):
        al.alSourceRewind(self.source)

#pause player
    def pause(self):
        al.alSourcePause(self.source)

#delete sound source
    def delete(self):
        al.alDeleteSources(1, self.source)

#Go straight to a set point in the sound file
    def _set_seek(self,offset):#float 0.0-1.0
        al.alSourcei(self.source,al.AL_BYTE_OFFSET,int(self.queue[0].length * offset))

#returns current buffer length position (IE: 21000), so divide by the buffers self.length
    def _get_seek(self):#returns float 0.0-1.0
        al.alGetSourcei(self.source, al.AL_BYTE_OFFSET, self.state)
        return float(self.state.value)/float(self.queue[0].length)

    rolloff = property(_get_rolloff, _set_rolloff,doc="""get/set rolloff factor""")
    volume = property(_get_volume, _set_volume,doc="""get/set volume""")
    pitch = property(_get_pitch, _set_pitch, doc="""get/set pitch""")
    loop = property(_get_loop, _set_loop, doc="""get/set loop state""")
    position = property(_get_position, _set_position,doc="""get/set position""")
    seek = property(_get_seek, _set_seek, doc="""get/set the current play position""")





if __name__ == '__main__':
    window = example()
    pyglet.app.run()
